
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <assert.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <stdlib.h>

#include <zend_extensions.h>
#include <zend_variables.h>
#include <zend_hash.h>
#include <php.h>
#include <php_ini.h>
#include <php_globals.h>
#include <SAPI.h>

#include "ftrace.h"

#ifdef ZTS
//////////////////////////////////////////////////////////////////////////////
//
int ftrace_globals_id;
//
////
#else
////
//
php_ftrace_globals ftrace_globals;
//
//////////////////////////////////////////////////////////////////////////////
#endif

//
// Create and export a zend_extension_version_info
//
ZEND_EXTENSION();

//
// Create and export a zend_extension.
//
ZEND_EXT_API zend_extension zend_extension_entry_suds =
{
    "Suds wuz here!",
    "1.1.1",
    "robert@webmotion.com",
    "http://www.webmotion.com",
    "Webmotion Inc.",

    NULL,
    NULL,
    ftraceActivate,
    NULL,
    NULL,
    NULL,
    NULL,
    onFCallBegin,
    onFCallEnd,
    NULL,
    NULL,

    STANDARD_ZEND_EXTENSION_PROPERTIES
};

ZEND_EXT_API void ftraceActivate()
{
    char szCommand[128];
    zval tempVal;

    CLS_FETCH();
    ELS_FETCH();
    SLS_FETCH();

    CG( extended_info ) = 1;

    sprintf( szCommand, "$GLOBALS['functionStack'] = array()" );

    ZEND_EVAL_STRING( szCommand, &tempVal );
}

#define rFLD struct robFunctionLinkDef

struct robFunctionLinkDef
{
    char * callingFile;
    char * functionName;
    long   lineNumber;

    struct robFunctionLinkDef * nextLink;
};

rFLD * createRobFunctionLink()
{
    rFLD * newLink = NULL;

    if( (newLink = (rFLD *)emalloc( sizeof( rFLD ) )) )
    {
        newLink->callingFile  = NULL;
        newLink->functionName = NULL;
        newLink->lineNumber   = 0L;

        newLink->nextLink     = NULL;
    }

    return newLink;
}

rFLD * deleteRobFunctionLink( rFLD * aLink )
{
    if( aLink->callingFile )
    {
        efree( aLink->callingFile );
    }

    if( aLink->functionName )
    {
        efree( aLink->functionName );
    }

    efree( aLink );

    return NULL;
}

rFLD * robFunctionStack = NULL;

void pushRobFunctionStats( char * callingFile, char * functionName, long lineNumber )
{
    rFLD * newLink;

    newLink = createRobFunctionLink();

    newLink->callingFile  = estrdup( callingFile );
    newLink->functionName = estrdup( functionName );
    newLink->lineNumber   = lineNumber;

    newLink->nextLink = robFunctionStack;
    robFunctionStack = newLink;
}

void popRobFunctionStats( void )
{
    rFLD * newBase = NULL;

    if( robFunctionStack )
    {
        newBase = robFunctionStack->nextLink;
    }

    deleteRobFunctionLink( robFunctionStack );
    robFunctionStack = newBase;
}

ZEND_EXT_API void onFCallBegin( zend_op_array* op_array )
{
    static char callFile[1024],
                callFunc[1024],
                tBuff[1024];

    int i, j;

    CLS_FETCH();
    ELS_FETCH();
    SLS_FETCH();

    sprintf( tBuff, "%s", zend_get_executed_filename()
                          ? zend_get_executed_filename()
                          : "(null)" );
    for( i = 0, j = 0; i < strlen( tBuff ), j < 1023; i++ )
    {
        if( tBuff[i] == '\'' )
        {
            callFile[j++] = '\\';
        }

        callFile[j++] = tBuff[i];
    }
    callFile[j] = 0x00;

    sprintf( tBuff, "%s", get_active_function_name()
                          ? get_active_function_name()
                          : "(null)" );
    for( i = 0, j = 0; i < strlen( tBuff ), j < 1023; i++ )
    {
        if( tBuff[i] == '\'' )
        {
            callFunc[j++] = '\\';
        }

        callFunc[j++] = tBuff[i];
    }
    callFunc[j] = 0x00;

    pushRobFunctionStats( callFile, callFunc, zend_get_executed_lineno() );
}

ZEND_EXT_API void onFCallEnd( zend_op_array* op_array )
{
    popRobFunctionStats();
}

//
// Function prototype declarations.
//
ZEND_FUNCTION( confirm_ftrace_compiled );
ZEND_FUNCTION( get_function_call_stack );

//
// Compiled function list so Zend knows what's in this module.
//
zend_function_entry ftrace_functions[] =
{
    ZEND_FE( confirm_ftrace_compiled, NULL )
    ZEND_FE( get_function_call_stack, NULL )
    {NULL, NULL, NULL}
};

//
// Compiled module information.
//
zend_module_entry ftrace_module_entry =
{
#if ZEND_MODULE_API_NO >= 20010901
    STANDARD_MODULE_HEADER,
#endif
    "ftrace",
    ftrace_functions,
    PHP_MINIT( ftrace ),
    PHP_MSHUTDOWN( ftrace ),
    PHP_RINIT( ftrace ),
    PHP_RSHUTDOWN( ftrace ),
    PHP_MINFO( ftrace ),
#if ZEND_MODULE_API_NO >= 20010901
    "Beats Me!",
#endif
    STANDARD_MODULE_PROPERTIES
};

#if COMPILE_DL_FTRACE
//////////////////////////////////////////////////////////////////////////////
//
// Implement "stub" routine so Zend knows who we are.
//
ZEND_GET_MODULE( ftrace )
//
//////////////////////////////////////////////////////////////////////////////
#endif

PHP_MINIT_FUNCTION(ftrace)
{
/* Remove comments if you have entries in php.ini
    REGISTER_INI_ENTRIES();
*/
    zend_register_extension( &zend_extension_entry_suds, (DL_HANDLE) 0);
    return SUCCESS;
}

PHP_MSHUTDOWN_FUNCTION(ftrace)
{
/* Remove comments if you have entries in php.ini
    UNREGISTER_INI_ENTRIES();
*/
    return SUCCESS;
}

//
// Routine to be called when phpinfo() is called.
//
PHP_MINFO_FUNCTION(ftrace)
{
	php_info_print_table_start();
	php_info_print_table_header(2, "ftrace support", "enabled");
	php_info_print_table_end();

	/* Remove comments if you have entries in php.ini
	DISPLAY_INI_ENTRIES();
	*/
}

//
// Remove if there's nothing to do at request start.
//
PHP_RINIT_FUNCTION(ftrace)
{
    return SUCCESS;
}

//
// Remove if there's nothing to do at request end.
//
PHP_RSHUTDOWN_FUNCTION(ftrace)
{
    while( robFunctionStack )
    {
        popRobFunctionStats();
    }

    return SUCCESS;
}

ZEND_FUNCTION( confirm_ftrace_compiled )
{
	zval **arg;
	int len;
	char string[256];

	if (ZEND_NUM_ARGS() != 1 || zend_get_parameters_ex(1, &arg) == FAILURE)
	{
		WRONG_PARAM_COUNT;
	}

	convert_to_string_ex( arg );

	len = sprintf( string, "Congratulations, you have successfully modified ext/ftrace/config.m4, module %s is compiled into PHP", Z_STRVAL_PP(arg) );

	RETVAL_STRINGL( string, len, 1 );
}

ZEND_FUNCTION( get_function_call_stack )
{
    zval *aHash;

    rFLD *currLink;

    array_init( return_value );

    currLink = robFunctionStack;
    while( currLink )
    {
        MAKE_STD_ZVAL( aHash );
        array_init( aHash ); 

        add_assoc_string( aHash, "callingFile", currLink->callingFile, 1 );
        add_assoc_string( aHash, "callingFunction", currLink->functionName, 1 );
        add_assoc_long( aHash, "lineNumber", currLink->lineNumber );

        add_next_index_zval( return_value, aHash );

        currLink = currLink->nextLink;
    }
}

